#! /bin/bash

# 2005-03-20	Begin.
# 2005-03-21	Last modif.

srcdir=../src
src_list="$srcdir/dtostre.c $srcdir/ftoa_engine.S"
#src_list="-lm"
avr_gcc=avr-gcc
#avr_gcc=~/usr/gcc-3.4.3/bin/avr-gcc

set -e

cflags="-Wall -W -Wundef -Werror -I$srcdir"
core=core_avr_dump.core

simulavr=simulavr
if ! $simulavr --version 2>/dev/null ; then
    simulavr=~/usr/simulavr/bin/simulavr
fi

simul ()
{
    echo "-------- $title:  $*"
    mcu=$1
    shift
    $avr_gcc -W -Wall -I. -mmcu=$mcu $* -o $bname.elf $bname.c $src_list
    exit_addr=0x`avr-nm $bname.elf | grep __stop_program | cut -f1 -d' '`
    avr-objcopy -O binary $bname.elf $bname.bin
    $simulavr -d $mcu -B $exit_addr -C $bname.bin >/dev/null
    if ! grep -q r24=00 $core || ! grep -q r25=00 $core
    then
	echo >&2 "*** Simulate: exit with non-zero code"
	exit 1
    fi
    rm -f $bname.{elf,bin} $core
}

run_sim ()
{
    for mcu in at90s8515 atmega8 ; do
	for opt in "-Os -frename-registers" ; do
	    simul $mcu $opt
	done
    done
}

measure_size ()
{
    echo "-------- size:  $*"
    bname=`basename $1 .c`
    shift
    mcu=$1
    shift
    $avr_gcc $cflags -mmcu=$mcu $* -o $bname.elf $bname.c $src_list
    avr-size $bname.elf
    rm -f $bname.elf
}

run_size ()
{
    for mcu in at90s8515 atmega8 ; do
	for opt in "-Os -frename-registers" ; do
	    measure_size $1 $mcu $opt
	done
    done
}

measure_speed ()
{
    echo "-------- speed:  $*"
    bname=`basename $1 .c`
    shift
    mcu=$1
    shift
    for nloops in 10 11 ; do
        $avr_gcc $cflags -mmcu=$mcu $* -DNLOOPS=$nloops \
	    -o $bname.elf $bname.c $src_list
	exit_addr=0x`avr-nm $bname.elf | grep __stop_program | cut -f1 -d' '`
        avr-objcopy -O binary $bname.elf $bname.bin
	$simulavr -d $mcu -B $exit_addr -C $bname.bin | grep "clock cycles"
        if ! grep -q r24=00 $core || ! grep -q r25=00 $core
	then
	    echo >&2 "*** Simulate: exit with non-zero code"
	    exit 1
	fi
    done
    rm -f $bname.{elf,bin} $core
}

run_speed ()
{
    for mcu in at90s8515 atmega8 ; do
	for opt in "-Os -frename-registers" ; do
	    measure_speed $1 $mcu $opt
	done
    done
}

run_glibc ()
{
    echo "-------- $title:  test of test"
    cc -W -Wall -Werror -D TEST_OF_TEST -Os -o $bname.exe $bname.c
    ./$bname.exe || {
	echo >&2 "*** Run test of test: exit with $?" ;
        exit 1 ;
    }
    rm -f $bname.exe
}


for f in tsize_*.c ; do
    run_size $f
done

for f in tspeed_*.c ; do
    run_speed $f
done

echo "All OK"
# eof
